#include <winresrc.h>

VS_VERSION_INFO VERSIONINFO
	FILEVERSION 0,2,1,0
	PRODUCTVERSION 0,2,1,0
	FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
	FILEFLAGS 0x21L
#else
	FILEFLAGS 0x20L
#endif
	FILEOS 0x40004L
	FILETYPE 0x1L
	FILESUBTYPE 0x0L
BEGIN
	BLOCK "StringFileInfo"
	BEGIN
		BLOCK "040904b0"
		BEGIN
			VALUE "Comments", "Provided under the terms of the GNU General Public License (LGPLv2.1+).\0"
			VALUE "CompanyName", "OpenSC Project\0"
			VALUE "FileDescription", "OpenSSL PKCS#11 engine\0"
			VALUE "FileVersion", "0,2,1,0\0"
			VALUE "InternalName", "engine_pkcs11\0"
			VALUE "LegalCopyright", "OpenSC Project\0"
			VALUE "LegalTrademarks", "\0"
			VALUE "OriginalFilename", "engine_pkcs11.dll\0"
			VALUE "PrivateBuild", "\0"
			VALUE "ProductName", "engine_pkcs11\0"
			VALUE "ProductVersion", "0,2,1,0\0"
			VALUE "SpecialBuild", "\0"
		END
	END
END

